from typing import List, Optional

from pydantic import BaseModel, Field  # pylint: ignore, Field


class _GetStranger_Resp(BaseModel):
    user_id: str
    nickname: str
    age: Optional[int] = None

    class Config:
        extra = "ignore"


class _GetFriends_Resp_FriendItem(BaseModel):
    user_id: str
    nickname: str
    remark: str


class _GetFriends_Resp(BaseModel):
    __root__: List[_GetFriends_Resp_FriendItem]

    class Config:
        extra = "ignore"


class _GetGroups_Resp_GroupItem(BaseModel):
    group_id: str
    group_name: str
    member_count: int
    max_member_count: int


class _GetGroups_Resp(BaseModel):
    __root__: List[_GetGroups_Resp_GroupItem]

    class Config:
        extra = "ignore"


class _GetMembers_Resp_MemberItem(BaseModel):
    group_id: str
    user_id: str
    name: str = Field(..., alias="nickname")
    nickname: str = Field(..., alias="card")
    role: str
    title: str

    class Config:
        extra = "ignore"


class _GetMembers_Resp(BaseModel):
    __root__: List[_GetMembers_Resp_MemberItem]

    class Config:
        extra = "ignore"



from typing import (TYPE_CHECKING, Any, Awaitable, Callable, Dict, Optional,
                    Tuple)

from graia.broadcast.entities.event import Dispatchable

from avilla.core.contactable import Contactable
from avilla.core.event.notice import (FileInfo, FriendAdd, FriendRevoke,
                                      GroupFileUploadNotice, GroupRevoke,
                                      MemberDemotedFromAdministrator,
                                      MemberJoinedByApprove,
                                      MemberJoinedByInvite, MemberLeave,
                                      MemberMuted,
                                      MemberPromotedToAdministrator,
                                      MemberRemoved, MemberUnmuted)
from avilla.core.event.request import FriendAddRequest, GroupJoinRequest
from avilla.core.event.service import (NetworkConnected, ServiceOffline,
                                       ServiceOnline)
from avilla.core.execution.fetch import (FetchFriend, FetchGroup, FetchMember,
                                         FetchStranger)
from avilla.onebot.event import HeartbeatReceived, NudgeEvent

"""
from avilla.event.message import *
from avilla.event.notice import *
from avilla.event.request import *
"""
from datetime import datetime, timedelta

from avilla.core.builtins.profile import (FriendProfile, GroupProfile,
                                          MemberProfile)
from avilla.core.event.message import MessageEvent
from avilla.core.role import Role
from avilla.onebot.profile import AnonymousProfile

if TYPE_CHECKING:
    from avilla.onebot.protocol import OnebotProtocol

ROLE_MAP = {"owner": Role.Owner, "admin": Role.Admin, "member": Role.Member}

# (post_type, sub_type=None, msg_type=None, notice_type=None, request_type=None, meta_event_type=None)\
#  => Type[Dispatchable]
EVENT_PARSING_TREE: Dict[
    Tuple[Optional[str], Optional[str], Optional[str], Optional[str], Optional[str], Optional[str]],
    Callable[["OnebotProtocol", Dict[str, Any]], Awaitable[Dispatchable]],
] = {}


def gen_parsing_key(data: Dict[str, str]):
    return (
        data.get("post_type"),
        data.get("sub_type"),
        data.get("message_type"),
        data.get("notice_type"),
        data.get("request_type"),
        data.get("meta_event_type"),
    )


def register(post, msg=None, sub=None, notice_type=None, request_type=None, meta_event_type=None):
    def wrapper(func: Callable[["OnebotProtocol", Dict[str, Any]], Awaitable[Dispatchable]]):
        EVENT_PARSING_TREE[(post, sub, msg, notice_type, request_type, meta_event_type)] = func
        return func

    return wrapper


@register("message", "private", "friend")
async def message_private_friend(proto: "OnebotProtocol", data: Dict[str, Any]):
    return MessageEvent(
        ctx=Contactable(
            data["user_id"],
            FriendProfile(name=data["sender"]["nickname"], remark=data["sender"]["nickname"]),
        ),
        message=await proto.parse_message(data["message"]),
        message_id=str(data["message_id"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register("message", "private", "group")
async def message_private_group(proto: "OnebotProtocol", data: Dict[str, Any]):
    return MessageEvent(
        ctx=Contactable(
            data["user_id"],
            await proto.ensure_execution(FetchMember(data["group_id"], data["user_id"])),
        ),
        message=await proto.parse_message(data["message"]),
        message_id=str(data["message_id"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register("message", "group", "normal")
async def message_group_normal(proto: "OnebotProtocol", data: Dict[str, Any]):
    return MessageEvent(
        ctx=Contactable(
            data["user_id"],
            MemberProfile(
                name=data["sender"]["nickname"],
                group=await proto.ensure_execution(FetchGroup(data["group_id"])),
                nickname=data["sender"]["card"],
                role=ROLE_MAP[data["sender"]["role"]],
                title=data["sender"]["title"],
            ),
        ),
        message=await proto.parse_message(data["message"]),
        message_id=str(data["message_id"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register("message", "group", "anonymous")
async def message_group_anonymous(proto: "OnebotProtocol", data: Dict[str, Any]):
    return MessageEvent(
        ctx=Contactable(
            data["user_id"],
            AnonymousProfile(
                id=data["anonymous"]["id"],
                group=await proto.ensure_execution(FetchGroup(data["group_id"])),
                name=data["anonymous"]["name"],
                _internal_id=data["anonymous"]["flag"],
            ),
        ),
        message=await proto.parse_message(data["message"]),
        message_id=str(data["message_id"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_upload")
async def notice_group_upload(proto: "OnebotProtocol", data: Dict[str, Any]):
    return GroupFileUploadNotice(
        ctx=await proto.ensure_execution(FetchMember(str(data["group_id"]), data["user_id"])),
        file=FileInfo(data["file"]["id"], data["file"]["name"], data["file"]["size"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_admin", sub="set")
async def notice_group_admin_set(protocol: "OnebotProtocol", data: Dict):
    return MemberPromotedToAdministrator(
        ctx=await protocol.ensure_execution(FetchMember(str(data["group_id"]), data["user_id"])),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_admin", sub="unset")
async def notice_group_admin_unset(protocol: "OnebotProtocol", data: Dict):
    return MemberDemotedFromAdministrator(
        ctx=await protocol.ensure_execution(FetchMember(str(data["group_id"]), data["user_id"])),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_decrease", sub="leave")
async def notice_group_decrease_leave(protocol: "OnebotProtocol", data: Dict):
    return MemberLeave(
        ctx=Contactable(
            id=str(data["user_id"]),
            profile=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        ),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_decrease", sub="kick")
async def notice_group_decrease_kick(protocol: "OnebotProtocol", data: Dict):
    return MemberRemoved(
        ctx=Contactable(
            id=str(data["user_id"]),
            profile=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        ),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_decrease", sub="kick_me")
async def notice_group_decrease_kick_me(protocol: "OnebotProtocol", data: Dict):
    g = Contactable(id=data["group_id"], profile=GroupProfile())
    return MemberRemoved(
        ctx=protocol.get_self(),
        group=g,
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_increase", sub="approve")
async def notice_group_increase_approve(protocol: "OnebotProtocol", data: Dict):
    return MemberJoinedByApprove(
        ctx=await protocol.ensure_execution(FetchMember(str(data["group_id"]), data["user_id"])),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        operator=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), data["operator_id"])
        ),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_increase", sub="invite")
async def notice_group_increase_invite(protocol: "OnebotProtocol", data: Dict):
    return MemberJoinedByInvite(
        ctx=await protocol.ensure_execution(FetchMember(str(data["group_id"]), data["user_id"])),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        operator=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), data["operator_id"])
        ),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_ban", sub="ban")
async def notice_group_ban(protocol: "OnebotProtocol", data: Dict):
    return MemberMuted(
        ctx=await protocol.ensure_execution(FetchMember(str(data["group_id"]), data["user_id"])),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        operator=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), data["operator_id"])
        ),
        duration=timedelta(seconds=data["duration"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_ban", sub="lift_ban")
async def notice_group_ban_lift(protocol: "OnebotProtocol", data: Dict):
    return MemberUnmuted(
        ctx=await protocol.ensure_execution(FetchMember(str(data["group_id"]), data["user_id"])),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        operator=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), data["operator_id"])
        ),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="friend_add")
async def notice_friend_add(protocol: "OnebotProtocol", data: Dict):
    return FriendAdd(
        ctx=await protocol.ensure_execution(FetchFriend(str(data["user_id"]))),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="group_recall")
async def notice_group_recall(protocol: "OnebotProtocol", data: Dict):
    return GroupRevoke(
        ctx=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), str(data["user_id"]))
        ),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        operator=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), data["operator_id"])
        ),
        message_id=str(data["message_id"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="notice", notice_type="friend_recall")
async def notice_friend_recall(protocol: "OnebotProtocol", data: Dict):
    return FriendRevoke(
        ctx=await protocol.ensure_execution(FetchFriend(data["user_id"])),
        message_id=str(data["message_id"]),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="request", request_type="friend")
async def request_friend(protocol: "OnebotProtocol", data: Dict):
    return FriendAddRequest(
        ctx=await protocol.ensure_execution(FetchStranger(str(data["user_id"]))),
        comment=data["comment"],
        request_id=data["flag"],
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="request", request_type="group", sub="add")
async def request_group_add(protocol: "OnebotProtocol", data: Dict):
    return GroupJoinRequest(
        ctx=await protocol.ensure_execution(FetchStranger(str(data["user_id"]))),
        comment=data["comment"],
        request_id=data["flag"],
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="request", request_type="group", sub="invite")
async def request_group_invite(protocol: "OnebotProtocol", data: Dict):
    return GroupJoinRequest(
        ctx=await protocol.ensure_execution(FetchStranger(str(data["user_id"]))),
        comment=data["comment"],
        is_invite=True,
        request_id=data["flag"],
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )


@register(post="meta_event", meta_event_type="lifecycle", sub="enable")
async def meta_event_lifecycle_enable(protocol: "OnebotProtocol", data: Dict):
    return ServiceOnline()


@register(post="meta_event", meta_event_type="lifecycle", sub="disable")
async def meta_event_lifecycle_disable(protocol: "OnebotProtocol", data: Dict):
    return ServiceOffline()


@register(post="meta_event", meta_event_type="lifecycle", sub="connect")
async def meta_event_lifecycle_connect(protocol: "OnebotProtocol", data: Dict):
    return NetworkConnected()


@register(post="meta_event", meta_event_type="heartbeat")
async def meta_event_heartbeat(protocol: "OnebotProtocol", data: Dict):
    return HeartbeatReceived(data["status"], data["interval"])


@register(post="notice", notice_type="notify", sub="poke")
async def notice_notify_poke(protocol: "OnebotProtocol", data: Dict):
    return NudgeEvent(
        ctx=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), str(data["target_id"]))
        ),
        group=await protocol.ensure_execution(FetchGroup(str(data["group_id"]))),
        operator=await protocol.ensure_execution(
            FetchMember(str(data["group_id"]), data["user_id"])
        ),
        current_id=str(data["self_id"]),
        time=datetime.fromtimestamp(data["time"]),
    )



from typing import Optional

from pydantic import BaseModel  # pylint: ignore

from avilla.core.builtins.profile import GroupProfile
from avilla.core.contactable import Contactable
from avilla.core.profile import BaseProfile


class AnonymousProfile(BaseModel, BaseProfile):
    id: str
    name: str
    group: Contactable[GroupProfile]

    _internal_id: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True



import base64

from avilla.core.builtins.elements import (Image, Notice, NoticeAll, Text,
                                           Video, Voice)
from avilla.core.utilles.message import MessageSerializeBus
from avilla.onebot.elements import (Anonymous, CustomMusicShare, Dice, Face,
                                    FlashImage, FriendRecommend,
                                    GroupRecommend, JsonMessage, Location,
                                    MusicShare, Poke, Rps, Shake, Share,
                                    XmlMessage)

onebot_msg_serializer = MessageSerializeBus()


@onebot_msg_serializer.register(Text)
async def _text_serializer(element: Text):
    return {"type": "text", "data": {"text": element.text}}


@onebot_msg_serializer.register(Image)
async def _image_serializer(element: Image):
    return {
        "type": "image",
        "data": {"file": f"base64://{base64.b64encode(await element.provider()).decode('utf-8')}"},
    }


@onebot_msg_serializer.register(Voice)
async def _voice_serializer(element: Voice):
    return {
        "type": "record",
        "data": {"file": f"base64://{base64.b64encode(await element.provider()).decode('utf-8')}"},
    }


@onebot_msg_serializer.register(Notice)
async def _notice_serializer(element: Notice):
    return {"type": "at", "data": {"qq": int(element.target)}}


@onebot_msg_serializer.register(NoticeAll)
async def _noticeall_serializer(_: NoticeAll):
    return {"type": "at", "data": {"qq": "all"}}


@onebot_msg_serializer.register(Video)
async def _video_serializer(element: Video):
    return {
        "type": "video",
        "data": {"file": f"base64://{base64.b64encode(await element.provider()).decode('utf-8')}"},
    }


@onebot_msg_serializer.register(FlashImage)
async def _flashimage_serializer(element: FlashImage):
    return {
        "type": "image",
        "data": {
            "type": "flash",
            "file": f"base64://{base64.b64encode(await element.provider()).decode('utf-8')}",
        },
    }


@onebot_msg_serializer.register(Face)
async def _face_serializer(element: Face):
    return {"type": "face", "data": {"id": element.id}}


@onebot_msg_serializer.register(Rps)
async def _rps_serializer(_: Rps):
    return {"type": "rps", "data": {}}


@onebot_msg_serializer.register(Dice)
async def _dice_serializer(_: Dice):
    return {"type": "dice", "data": {}}


@onebot_msg_serializer.register(Shake)
async def _shake_serializer(_: Shake):
    return {"type": "shake", "data": {}}


@onebot_msg_serializer.register(Poke)
async def _poke_serializer(element: Poke):
    return {
        "type": "poke",
        "data": {
            "type": element.type,
            "id": element.id,
        },
    }


@onebot_msg_serializer.register(Anonymous)
async def _anonymous_serializer(_: Anonymous):
    return {"type": "anonymous", "data": {}}


@onebot_msg_serializer.register(Share)
async def _share_serializer(element: Share):
    return {
        "type": "share",
        "data": {
            "url": element.url,
            "title": element.title,
            "content": element.content,
            "image": element.image,
        },
    }


@onebot_msg_serializer.register(FriendRecommend)
async def _friendrecommend_serializer(element: FriendRecommend):
    return {"type": "contact", "data": {"type": "qq", "id": element.id}}


@onebot_msg_serializer.register(GroupRecommend)
async def _grouprecommend_serializer(element: GroupRecommend):
    return {"type": "contact", "data": {"type": "group", "id": element.id}}


@onebot_msg_serializer.register(Location)
async def _location_serializer(element: Location):
    return {
        "type": "location",
        "data": {
            "lat": element.lat,
            "lon": element.lon,
            "title": element.title,
            "content": element.content,
        },
    }


@onebot_msg_serializer.register(MusicShare)
async def _musicshare_serializer(element: MusicShare):
    return {"type": "music", "data": {"type": element.type, "id": element.id}}


@onebot_msg_serializer.register(CustomMusicShare)
async def _custommusicshare_serializer(element: CustomMusicShare):
    return {
        "type": "music",
        "data": {
            "type": "custom",
            "url": element.url,
            "audio": element.audio,
            "title": element.title,
            "content": element.content,
            "image": element.image,
        },
    }


@onebot_msg_serializer.register(XmlMessage)
async def _xmlmessage_serializer(element: XmlMessage):
    return {"type": "xml", "data": {"xml": element.xml}}


@onebot_msg_serializer.register(JsonMessage)
async def _jsonmessage_serializer(element: JsonMessage):
    return {"type": "json", "data": {"json": element.json}}



from typing import Literal, Optional

from avilla.core.builtins.elements import Image
from avilla.core.message.chain import MessageChain
from avilla.core.message.element import Element

__all__ = (
    "FlashImage",
    "Face",
    "Rps",
    "Dice",
    "Shake",
    "Poke",
    "Anonymous",
    "Share",
    "FriendRecommend",
    "GroupRecommend",
    "Location",
    "MusicShare",
    "CustomMusicShare",
    "MergedForward",
    "MergedForwardNode",
    "MergedForwardCustomNode",
    "XmlMessage",
    "JsonMessage",
)


class FlashImage(Image):
    def asDisplay(self) -> str:
        return "[$onebot::FlashImage]"


class Face(Element):
    id: int

    def __init__(self, id: int) -> None:
        self.id = id

    def asDisplay(self) -> str:
        return f"[$onebot::Face:id={self.id}]"


class Rps(Element):
    def asDisplay(self) -> str:
        return "[$onebot::Rps]"


class Dice(Element):
    def asDisplay(self) -> str:
        return "[$onebot::Dice]"


class Shake(Element):
    def asDisplay(self) -> str:
        return "[$onebot::Shake]"


class Poke(Element):
    type: str
    id: str
    name: Optional[str] = None

    def __init__(self, type: str, id: str, name: Optional[str] = None) -> None:
        self.type = type
        self.id = id
        self.name = name

    def asDisplay(self) -> str:
        return f"[$onebot::Poke:type={self.type},id={self.id},name={self.name or 'null'}]"


class Anonymous(Element):
    ignore: bool = True

    def __init__(self, ignore: bool = True) -> None:
        self.ignore = ignore

    def asDisplay(self) -> str:
        return "[$onebot::#Anonymous#]"


class Share(Element):
    url: str
    title: str
    content: Optional[str] = None
    image: Optional[str] = None

    def __init__(
        self, url: str, title: str, content: Optional[str] = None, image: Optional[str] = None
    ) -> None:
        self.url = url
        self.title = title
        self.content = content
        self.image = image

    def asDisplay(self) -> str:
        return f"[$onebot::Share:\
            url={self.url},\
            title={self.title},\
            content={self.content or 'null'},\
            image={self.image or 'null'}]"


class FriendRecommend(Element):
    id: str

    def __init__(self, id: str) -> None:
        self.id = id

    def asDisplay(self) -> str:
        return f"[$onebot::FriendRecommend:id={self.id}]"


class GroupRecommend(Element):
    id: str

    def __init__(self, id: str) -> None:
        self.id = id

    def asDisplay(self) -> str:
        return f"[$onebot::GroupRecommend:id={self.id}]"


class Location(Element):
    lat: float
    lon: float
    title: Optional[str] = None
    content: Optional[str] = None

    def __init__(
        self, lat: float, lon: float, title: Optional[str] = None, content: Optional[str] = None
    ) -> None:
        self.lat = lat
        self.lon = lon
        self.title = title
        self.content = content

    def asDisplay(self) -> str:
        return f"[$onebot::Location:\
            lat={self.lat},\
            lon={self.lon},\
            title={self.title or 'null'},\
            content={self.content or 'null'}]"


class MusicShare(Element):
    type: Literal["qq", "163", "xm"]
    id: str

    def __init__(self, type: Literal["qq", "163", "xm"], id: str) -> None:
        self.type = type
        self.id = id

    def asDisplay(self) -> str:
        return f"[$onebot::MusicShare:type={self.type},id={self.id}]"


class CustomMusicShare(Element):
    url: str
    audio: str
    title: str
    content: Optional[str] = None
    image: Optional[str] = None

    def __init__(
        self,
        url: str,
        audio: str,
        title: str,
        content: Optional[str] = None,
        image: Optional[str] = None,
    ) -> None:
        self.url = url
        self.audio = audio
        self.title = title
        self.content = content
        self.image = image

    def asDisplay(self) -> str:
        return f"[$onebot::CustomMusicShare:\
            url={self.url},\
            audio={self.audio},\
            title={self.title},\
            content={self.content or 'null'},\
            image={self.image or 'null'}]"


class MergedForward(Element):
    id: str

    def __init__(self, id: str) -> None:
        self.id = id

    def asDisplay(self) -> str:
        return f"[$onebot::MergedForward:id={self.id}]"


class MergedForwardNode(Element):
    id: str

    def __init__(self, id: str) -> None:
        self.id = id

    def asDisplay(self) -> str:
        return f"[$onebot::MergedForwardNode:id={self.id}]"


class MergedForwardCustomNode(Element):
    user_id: str
    nickname: str
    content: MessageChain

    def __init__(self, user_id: str, nickname: str, content: MessageChain) -> None:
        self.user_id = user_id
        self.nickname = nickname
        self.content = content

    def asDisplay(self) -> str:
        return f"[$onebot::MergedForwardCustomNode:\
            user_id={self.user_id},\
            nickname={self.nickname},\
            content={self.content}]"


class XmlMessage(Element):
    xml: str

    def __init__(self, xml: str) -> None:
        self.xml = xml

    def asDisplay(self) -> str:
        return f"[$onebot::XmlMessage:xml={self.xml}]"


class JsonMessage(Element):
    json: str

    def __init__(self, _json: str) -> None:
        self.json = _json

    def asDisplay(self) -> str:
        return f"[$onebot::JsonMessage:json={self.json}]"



from avilla.core.builtins.elements import (Image, Notice, NoticeAll, Quote,
                                           Text, Video, Voice)
from avilla.core.builtins.profile import (FriendProfile, GroupProfile,
                                          MemberProfile, SelfProfile,
                                          StrangerProfile)
from avilla.core.event.notice import (FriendAdd, FriendRevoke,
                                      GroupFileUploadNotice, GroupRevoke,
                                      MemberDemotedFromAdministrator,
                                      MemberJoinedByApprove,
                                      MemberJoinedByInvite, MemberLeave,
                                      MemberMuted,
                                      MemberPromotedToAdministrator,
                                      MemberRemoved, MemberUnmuted)
from avilla.core.event.request import FriendAddRequest, GroupJoinRequest
from avilla.core.event.service import (NetworkConnected, ServiceOffline,
                                       ServiceOnline)
from avilla.onebot.elements import (Anonymous, Dice, Face, FriendRecommend,
                                    GroupRecommend, JsonMessage, Location,
                                    MergedForward, MergedForwardCustomNode,
                                    Poke, Rps, Shake, Share, XmlMessage)
from avilla.onebot.event import HeartbeatReceived, NudgeEvent

ABILITIES = {
    i.get_ability_id()
    for i in [
        FriendAdd,
        FriendRevoke,
        GroupFileUploadNotice,
        GroupRevoke,
        MemberDemotedFromAdministrator,
        MemberJoinedByApprove,
        MemberJoinedByInvite,
        MemberLeave,
        MemberMuted,
        MemberPromotedToAdministrator,
        MemberRemoved,
        MemberUnmuted,
        FriendAddRequest,
        GroupJoinRequest,
        NetworkConnected,
        ServiceOffline,
        ServiceOnline,
        HeartbeatReceived,
        NudgeEvent,
        Anonymous,
        Dice,
        Face,
        FriendRecommend,
        GroupRecommend,
        JsonMessage,
        Location,
        MergedForward,
        MergedForwardCustomNode,
        Poke,
        Rps,
        Shake,
        Share,
        XmlMessage,
        Image,
        Notice,
        NoticeAll,
        Text,
        Quote,
        Video,
        Voice,
        MemberProfile,
        GroupProfile,
        FriendProfile,
        StrangerProfile,
        SelfProfile,
    ]
}



from typing import Dict, Optional, Union

from pydantic import BaseModel, BaseSettings
from yarl import URL

from avilla.core.network.signatures import (ClientCommunicationMethod,
                                            ServiceCommunicationMethod)


class HttpCommunication(BaseModel, ClientCommunicationMethod):
    api_root: URL

    class Config:
        arbitrary_types_allowed = True


class ReverseHttpCommunication(BaseModel, ServiceCommunicationMethod):
    listening_host: str = "0.0.0.0"
    listening_port: int

    secret: Optional[str] = None
    timeout: int = -1


class WebsocketCommunication(BaseModel, ClientCommunicationMethod):
    api_root: URL

    class Config:
        arbitrary_types_allowed = True


class ReverseWebsocketCommunication(BaseModel, ServiceCommunicationMethod):
    listening_host: str = "0.0.0.0"
    listening_port: int


class OnebotConfig(BaseSettings):
    access_token: Optional[str] = None
    bot_id: str

    # 暂时只支持 aiohttp 的 client 模式(使用 websocket 获取事件什么的..)
    communications: Dict[
        str,
        Union[
            HttpCommunication,
            ReverseHttpCommunication,
            WebsocketCommunication,
            ReverseWebsocketCommunication,
        ],
    ]



import typing
from typing import Dict

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from avilla.core.builtins.profile import GroupProfile, MemberProfile
from avilla.core.contactable import Contactable
from avilla.core.event import AvillaEvent
from avilla.core.typing import T_Profile


class HeartbeatReceived(Dispatchable):
    status: Dict
    interval: int

    def __init__(self, status: Dict, interval: int):
        self.status = status
        self.interval = interval

    def __repr__(self) -> str:
        return f"<HeartbeatReceived: status={self.status} interval={self.interval})>"

    @staticmethod
    def get_ability_id() -> str:
        return "event::HeartbeatReceived"

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface):
            return


class NudgeEvent(AvillaEvent[T_Profile]):
    group: Contactable[GroupProfile]
    operator: Contactable[MemberProfile]

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: "DispatcherInterface[NudgeEvent]"):
            if (
                typing.get_origin(interface.annotation) is Contactable
                and (*typing.get_args(interface.annotation), None)[0] is GroupProfile
            ):
                return interface.event.group



from typing import Awaitable, Callable, Dict

from avilla.core.builtins.elements import (Image, Notice, NoticeAll, Quote,
                                           Text, Video, Voice)
from avilla.core.context import ctx_protocol
from avilla.core.message.element import Element
from avilla.core.provider import HttpGetProvider
from avilla.onebot.elements import (Anonymous, Dice, Face, FlashImage,
                                    FriendRecommend, GroupRecommend,
                                    JsonMessage, Location, MergedForward,
                                    MergedForwardCustomNode, Poke, Rps, Shake,
                                    Share, XmlMessage)

ELEMENT_TYPE_MAP: Dict[str, Callable[[Dict], Awaitable[Element]]] = {}


def register(element_type: str):
    def wrapper(func: Callable[[Dict], Awaitable[Element]]):
        ELEMENT_TYPE_MAP[element_type] = func
        return func

    return wrapper


@register("text")
async def text(data: Dict):
    return Text(data["text"])


@register("image")
async def image(data: Dict):
    if data.get("type") == "flash":
        return FlashImage(HttpGetProvider(data["url"]))
    return Image(HttpGetProvider(data["url"]))


@register("at")
async def at(data: Dict):
    if data["qq"] == "all":
        return NoticeAll()
    return Notice(str(data["qq"]))


@register("reply")
async def reply(data: Dict):
    return Quote(id=data["id"])


@register("record")
async def record(data: Dict):
    return Voice(HttpGetProvider(data["url"]))


@register("video")
async def video(data: Dict):
    return Video(HttpGetProvider(data["url"]))


@register("face")
async def face(data: Dict):
    return Face(data["id"])


@register("rps")
async def rps(data: Dict):
    return Rps()


@register("dice")
async def dice(data: Dict):
    return Dice()


@register("shake")
async def shake(data: Dict):
    return Shake()


@register("poke")
async def poke(data: Dict):
    return Poke(type=data["type"], id=data["id"], name=data["name"])


@register("anonymous")
async def anonymous(data: Dict):
    return Anonymous()


@register("share")
async def share(data: Dict):
    return Share(
        url=data["url"],
        title=data["title"],
        content=data["content"],
        image=data["image"],
    )


@register("contact")
async def contact(data: Dict):
    if data["type"] == "qq":
        return FriendRecommend(data["id"])
    return GroupRecommend(data["id"])


@register("location")
async def location(data: Dict):
    return Location(
        data["lat"],
        data["lon"],
        data["title"],
        data["content"],
    )


@register("forward")
async def forward(data: Dict):
    return MergedForward(
        data["id"],
    )


@register("node")
async def node(data: Dict):
    return MergedForwardCustomNode(
        data["user_id"], data["nickname"], await ctx_protocol.get().parse_message(data["content"])
    )


@register("xml")
async def xml(data: Dict):
    return XmlMessage(data["data"])


@register("json")
async def json(data: Dict):
    return JsonMessage(data["data"])



import asyncio
import contextlib
import json
import weakref
from typing import Any, Dict, Final, List, Tuple, Type, Union

from avilla.core import context
from avilla.core.builtins.profile import SelfProfile
from avilla.core.contactable import Contactable
from avilla.core.context import ctx_protocol
from avilla.core.execution import Execution
from avilla.core.message.chain import MessageChain
from avilla.core.network.client import (AbstractHttpClient,
                                        AbstractWebsocketClient, Client)
from avilla.core.network.service import Service
from avilla.core.network.signatures import (ClientCommunicationMethod,
                                            ServiceCommunicationMethod)
from avilla.core.platform import Platform
from avilla.core.protocol import BaseProtocol
from avilla.core.relationship import Relationship
from avilla.core.utilles import random_string
from avilla.core.utilles.transformer import (JsonTransformer,
                                             Utf8StringTransformer)
from avilla.onebot.ability import ABILITIES
from avilla.onebot.config import (HttpCommunication, OnebotConfig,
                                  ReverseWebsocketCommunication,
                                  WebsocketCommunication)
from avilla.onebot.element_tree import ELEMENT_TYPE_MAP

from .event_tree import EVENT_PARSING_TREE, gen_parsing_key
from .execution_ensure import ensure_execution
from .message_serializer import onebot_msg_serializer


class OnebotProtocol(BaseProtocol):
    config: OnebotConfig

    platform: Final[Platform] = Platform(
        name="Tencent QQ",
        protocol_provider_name="universal",
        implementation="OneBot",
        supported_impl_version="v11",
        generation="11",
    )

    _pending_futures: weakref.WeakValueDictionary[str, asyncio.Future]

    def __post_init__(self) -> None:
        self._pending_futures = weakref.WeakValueDictionary()

        if self.using_exec_method is WebsocketCommunication:
            ws_client: AbstractWebsocketClient = self.avilla.network_interface.get_network("ws")  # type: ignore
            ws_client.on_received_data(self.config.bot_id)(self._onebot_ws_data_received)

    def ensure_networks(
        self,
    ) -> Tuple[
        Dict[str, Union[Client, Service]],
        Union[Type[ClientCommunicationMethod], Type[ServiceCommunicationMethod]],
    ]:
        result = {}
        comm_method = None
        if "http" in self.config.communications:
            result["http"] = self.avilla.network_interface.get_network("http")
            comm_method = HttpCommunication
        if "http-service" in self.config.communications:
            result["http-service"] = self.avilla.network_interface.get_network("http-service")
        if "ws-service" in self.config.communications:
            result["ws-service"] = self.avilla.network_interface.get_network("ws-service")
            comm_method = ReverseWebsocketCommunication
        if "ws" in self.config.communications:
            result["ws"] = self.avilla.network_interface.get_network("ws")
            comm_method = WebsocketCommunication

        if result and comm_method:
            return result, comm_method
        else:
            raise TypeError("invaild config for network")

    def get_self(self) -> "Contactable[SelfProfile]":
        return Contactable(self.config.bot_id, SelfProfile())

    def has_ability(self, ability: str) -> bool:
        return ability in ABILITIES

    async def parse_message(self, data: List[Dict]) -> "MessageChain":
        result = []

        with ctx_protocol.use(self):
            for x in data:
                elem_type = x["type"]
                elem_parser = ELEMENT_TYPE_MAP.get(elem_type)
                if elem_parser:
                    result.append(await elem_parser(x["data"]))
                else:
                    self.avilla.logger.error(f"message_parser: unexpected element type {elem_type}")

        return MessageChain.create(result)

    async def serialize_message(self, message: "MessageChain") -> List:
        return await onebot_msg_serializer.serialize(message)

    async def get_relationship(self, entity: "Contactable") -> "Relationship":
        return await super().get_relationship(entity)

    async def launch_entry(self):
        comms = self.config.communications
        if self.using_exec_method is WebsocketCommunication:
            ws_client: AbstractWebsocketClient = self.using_networks["ws"]  # type: ignore
            ws_config: WebsocketCommunication = comms["ws"]  # type: ignore
            await ws_client.connect(
                ws_config.api_root, account=self.config.bot_id, headers=self._get_http_headers()
            )

    async def ensure_execution(self, execution: Execution) -> Any:
        return await ensure_execution(self, execution=execution)

    def _get_http_headers(self):
        return {
            "Content-Type": "application/json",
            **(
                {"Authorization": f"Bearer {self.config.access_token}"}
                if self.config.access_token
                else {}
            ),
        }

    async def _http_get(self, path: str, params: Dict[str, str] = None) -> Any:
        http_client: AbstractHttpClient = self.using_networks["http"]  # type: ignore
        return (
            (
                await http_client.get(
                    self.config.communications["http"]
                    .api_root.with_path(path)  # type:ignore
                    .with_query(params or {}),
                    headers=self._get_http_headers(),
                )
            )
            .passby(Utf8StringTransformer)
            .passby(JsonTransformer)
            .transform()
        )

    async def _http_post(self, path: str, data: Union[Dict, Any]) -> Any:
        http_client: AbstractHttpClient = self.using_networks["http"]  # type: ignore
        return (
            (
                await http_client.post(
                    self.config.communications["http"].api_root.with_path(path),  # type:ignore
                    json=data,
                    headers=self._get_http_headers(),
                )
            )
            .passby(Utf8StringTransformer)
            .passby(JsonTransformer)
            .transform()
        )

    async def _ws_client_send_packet(self, action: str, data: Any) -> Any:
        ws_client: AbstractWebsocketClient = self.using_networks["ws"]  # type: ignore

        response_id: str = random_string()

        ftr = asyncio.get_running_loop().create_future()
        self._pending_futures[response_id] = ftr
        await ws_client.send_json(
            self.config.bot_id, {"action": action, "params": data, "echo": response_id}
        )
        return await ftr

    async def _onebot_send_packet(self, endpoint: str, data: Any) -> Any:
        if self.using_exec_method is WebsocketCommunication:
            return await self._ws_client_send_packet(endpoint, data)
        elif self.using_exec_method is HttpCommunication:
            return await self._http_post("/" + endpoint, data)

    async def _onebot_ws_data_received(self, client, connid, raw_data: Union[str, bytes]):
        if isinstance(raw_data, bytes):
            raw_data = raw_data.decode()
        data: Dict = json.loads(raw_data)
        is_event = "post_type" in data
        if is_event:
            event_parser = EVENT_PARSING_TREE.get(gen_parsing_key(data))  # ignore: noqa
            if event_parser:
                event = await event_parser(self, data)
                if event.__class__.__name__ != "HeartbeatReceived":  # Heartbeat
                    self.avilla.logger.debug(
                        f"received event [{event.__class__.__name__}]: {repr(event)}"
                    )
                else:
                    self.avilla.logger.debug("received heartbeat packet")
                with contextlib.ExitStack() as stack:
                    stack.enter_context(context.ctx_avilla.use(self.avilla))
                    stack.enter_context(context.ctx_event.use(event))
                    stack.enter_context(context.ctx_protocol.use(self))
                    self.avilla.broadcast.postEvent(event)
            else:
                self.avilla.logger.error(
                    f"event receiver: unknown event: key={gen_parsing_key(data)}, data={data}"
                )
        else:
            p_ftr = self._pending_futures.get(data["echo"])
            if p_ftr:
                p_ftr.set_result(data)



from typing import TYPE_CHECKING, Any, Iterable

from yarl import URL

from avilla.core.builtins.profile import (FriendProfile, GroupProfile,
                                          MemberProfile, SelfProfile,
                                          StrangerProfile)
from avilla.core.builtins.resource import AvatarResource
from avilla.core.contactable import Contactable, ref
from avilla.core.context import ctx_target
from avilla.core.exceptions import ExecutionException
from avilla.core.execution.fetch import (FetchAvatar, FetchBot, FetchFriend,
                                         FetchFriends, FetchGroup, FetchGroups,
                                         FetchMember, FetchMembers,
                                         FetchStranger)
from avilla.core.execution.group import (GroupLeave, GroupMute, GroupNameSet,
                                         GroupUnmute,
                                         MemberDemoteFromAdministrator,
                                         MemberMute, MemberNicknameClear,
                                         MemberNicknameSet,
                                         MemberPromoteToAdministrator,
                                         MemberRemove, MemberSpecialTitleSet,
                                         MemberUnmute)
from avilla.core.execution.message import (MessageFetch, MessageFetchResult,
                                           MessageId, MessageRevoke,
                                           MessageSend, MessageSendPrivate)
from avilla.core.network.client import AbstractHttpClient
from avilla.core.provider import RawProvider
from avilla.core.role import Role
from avilla.core.utilles.override_bus import OverrideBus
from avilla.core.utilles.override_subbus import (execution_subbus,
                                                 network_method_subbus,
                                                 proto_ensure_exec_params)

from .resp import (_GetFriends_Resp, _GetFriends_Resp_FriendItem,
                   _GetGroups_Resp, _GetGroups_Resp_GroupItem,
                   _GetMembers_Resp, _GetMembers_Resp_MemberItem,
                   _GetStranger_Resp)

if TYPE_CHECKING:
    from .protocol import OnebotProtocol


ensure_execution: "OverrideBus[OnebotProtocol]" = OverrideBus(
    proto_ensure_exec_params,
    {"execution": execution_subbus, "network": network_method_subbus},
    {"network": lambda: "http"},
)


def _extract_and_check_as_groupid(target) -> int:
    if isinstance(target, str):
        if not target.isdigit():
            raise ValueError(f"invalid group id: {target}")
        return int(target)
    elif isinstance(target, int):
        return target
    elif isinstance(target, Contactable):
        if isinstance(target.profile, GroupProfile):
            if not target.id.isdigit():
                raise ValueError(f"invalid group id: {target.__repr__()}")
            return int(target.id)
        elif isinstance(target.profile, MemberProfile):
            if target.profile.group is None:
                raise ValueError(f"invalid input, maybe you can issue the problem: {target}")
            if not target.profile.group.id.isdigit():
                raise ValueError(f"invalid group id: {target.profile.group.id}")
            return int(target.profile.group.id)
        else:
            raise ValueError(f"invalid input, maybe you can issue the problem: {target}")
    elif isinstance(target, ref) and ref.type == "group":
        if not ref.id.isdigit():
            raise ValueError(f"invalid group id: {target.id}")
        return int(ref.id)
    else:
        raise ValueError(f"invalid target: {target}")


def _extract_and_check_as_memberid(target) -> int:
    if isinstance(target, str):
        if not target.isdigit():
            raise ValueError(f"invalid member id: {target}")
        return int(target)
    elif isinstance(target, int):
        return target
    elif isinstance(target, Contactable):
        return int(target.id)
    else:
        raise ValueError(f"invalid target: {target}")


def _check_execution(data: Any):
    if isinstance(data, dict):
        if data.get("status") == "failed":
            raise ExecutionException("execution failed")
        return data["data"]


@ensure_execution.override(execution=FetchBot, network="http")
@ensure_execution.override(execution=FetchBot, network="http-service")
@ensure_execution.override(execution=FetchBot, network="ws")
@ensure_execution.override(execution=FetchBot, network="ws-service")
async def get_bot(self: "OnebotProtocol", execution: FetchBot) -> "Contactable[SelfProfile]":
    return Contactable(self.config.bot_id, SelfProfile())


@ensure_execution.override(execution=FetchStranger, network="http")
async def get_stranger_http(self, execution: FetchStranger) -> "Contactable[StrangerProfile]":
    data = _GetStranger_Resp.parse_obj(
        _check_execution(
            await self._http_post("/get_stranger_info", {"user_id": int(execution.target)})
        )
    )

    return Contactable(id=data.user_id, profile=StrangerProfile(name=data.nickname, age=data.age))


@ensure_execution.override(execution=FetchStranger, network="ws")
async def get_stranger_ws(self, execution: FetchStranger) -> "Contactable[StrangerProfile]":
    data = _GetStranger_Resp.parse_obj(
        _check_execution(
            await self._ws_client_send_packet(
                "get_stranger_info", {"user_id": int(execution.target)}
            )
        )
    )

    return Contactable(id=data.user_id, profile=StrangerProfile(name=data.nickname, age=data.age))


@ensure_execution.override(execution=FetchFriend, network="http")
async def get_friend_http(self, execution: FetchFriend) -> "Contactable[FriendProfile]":
    data = _GetFriends_Resp_FriendItem.parse_obj(
        _check_execution(
            await self._http_post("/get_friend_info", {"user_id": int(execution.target)})
        )
    )

    return Contactable(
        id=data.user_id, profile=FriendProfile(name=data.nickname, remark=data.remark)
    )


@ensure_execution.override(execution=FetchFriend, network="ws")
async def get_friend_ws(
    self: "OnebotProtocol", execution: FetchFriend
) -> "Contactable[FriendProfile]":
    data = _GetFriends_Resp_FriendItem.parse_obj(
        _check_execution(
            await self._ws_client_send_packet("get_friend_info", {"user_id": int(execution.target)})
        )
    )

    return Contactable(
        id=data.user_id, profile=FriendProfile(name=data.nickname, remark=data.remark)
    )


@ensure_execution.override(execution=FetchFriends, network="http")
async def get_friends_http(self, execution: FetchFriends) -> "Iterable[Contactable[FriendProfile]]":
    data = _GetFriends_Resp.parse_obj(_check_execution(await self._http_get("/get_friends")))

    return [
        Contactable(id=i.user_id, profile=FriendProfile(name=i.nickname, remark=i.remark))
        for i in data.__root__
    ]


@ensure_execution.override(execution=FetchFriends, network="ws")
async def get_friends_ws(self, execution: FetchFriends) -> "Iterable[Contactable[FriendProfile]]":
    data = _GetFriends_Resp.parse_obj(
        _check_execution(await self._ws_client_send_packet("get_friends", {}))
    )

    return [
        Contactable(id=i.user_id, profile=FriendProfile(name=i.nickname, remark=i.remark))
        for i in data.__root__
    ]


@ensure_execution.override(execution=FetchGroup, netwotk="http")
async def get_group_http(
    self: "OnebotProtocol", execution: FetchGroup
) -> Contactable[GroupProfile]:
    group_id = _extract_and_check_as_groupid(execution.target)

    data = _GetGroups_Resp_GroupItem.parse_obj(
        _check_execution(
            await self._http_post(
                "/get_group_info",
                {"group_id": group_id},
            )
        )
    )

    return Contactable(
        id=data.group_id,
        profile=GroupProfile(
            name=data.group_name, counts=data.member_count, limit=data.max_member_count
        ),
    )


@ensure_execution.override(execution=FetchGroup, network="ws")
async def get_group_ws(self: "OnebotProtocol", execution: FetchGroup) -> Contactable[GroupProfile]:
    group_id = _extract_and_check_as_groupid(execution.target)

    data = _GetGroups_Resp_GroupItem.parse_obj(
        _check_execution(
            await self._ws_client_send_packet(
                "get_group_info",
                {"group_id": group_id},
            )
        )
    )

    return Contactable(
        id=data.group_id,
        profile=GroupProfile(
            name=data.group_name, counts=data.member_count, limit=data.max_member_count
        ),
    )


@ensure_execution.override(execution=FetchGroups, network="http")
async def get_groups_http(
    self: "OnebotProtocol", execution: FetchGroups
) -> "Iterable[Contactable]":
    data = _GetGroups_Resp.parse_obj(_check_execution(await self._http_get("/get_group_list")))

    return [
        Contactable(
            id=i.group_id,
            profile=GroupProfile(
                name=i.group_name, counts=i.member_count, limit=i.max_member_count
            ),
        )
        for i in data.__root__
    ]


@ensure_execution.override(execution=FetchGroups, network="ws")
async def get_groups_ws(self: "OnebotProtocol", execution: FetchGroups) -> "Iterable[Contactable]":
    data = _GetGroups_Resp.parse_obj(
        _check_execution(await self._ws_client_send_packet("get_group_list", {}))
    )

    return [
        Contactable(id=i.group_id, profile=GroupProfile(name=i.group_name)) for i in data.__root__
    ]


@ensure_execution.override(execution=FetchMembers, network="http")
async def get_members_http(self, execution: FetchMembers) -> "Iterable[Contactable[MemberProfile]]":
    group_id = _extract_and_check_as_groupid(execution.group)

    data = _GetMembers_Resp.parse_obj(
        _check_execution(
            await self._http_get(
                "/get_group_member_list",
                {"group_id": str(group_id)},
            )
        )
    )

    return [
        Contactable(
            id=i.user_id,
            profile=MemberProfile(
                name=i.name,
                role={
                    "owner": Role.Owner,
                    "admin": Role.Admin,
                    "member": Role.Member,
                }[i.role],
                nickname=i.nickname,
                title=i.title,
                group=Contactable(str(group_id), GroupProfile()),
            ),
        )
        for i in data.__root__
    ]


@ensure_execution.override(execution=FetchMembers, network="ws")
async def get_members_ws(self, execution: FetchMembers) -> "Iterable[Contactable[MemberProfile]]":
    group_id = _extract_and_check_as_groupid(execution.group)

    data = _GetMembers_Resp.parse_obj(
        _check_execution(
            await self._ws_client_send_packet(
                "get_group_member_list",
                {"group_id": group_id},
            )
        )
    )

    return [
        Contactable(
            i.user_id,
            MemberProfile(
                name=i.name,
                group=Contactable(str(group_id), GroupProfile()),
                role={
                    "owner": Role.Owner,
                    "admin": Role.Admin,
                    "member": Role.Member,
                }[i.role],
                nickname=i.nickname,
                title=i.title,
            ),
        )
        for i in data.__root__
    ]


@ensure_execution.override(execution=FetchMember, network="http")
@ensure_execution.override(execution=FetchMember, network="ws")
async def get_member_http(
    self: "OnebotProtocol", execution: FetchMember
) -> "Contactable[MemberProfile]":
    group_id = _extract_and_check_as_groupid(execution.group)

    data = _GetMembers_Resp_MemberItem.parse_obj(
        _check_execution(
            await self._onebot_send_packet(
                "get_group_member_info",
                {
                    "group_id": group_id,
                    "user_id": int(execution.target),
                },
            )
        )
    )

    return Contactable(
        data.user_id,
        MemberProfile(
            name=data.name,
            group=Contactable(str(group_id), GroupProfile()),
            role={
                "owner": Role.Owner,
                "admin": Role.Admin,
                "member": Role.Member,
            }[data.role],
            nickname=data.nickname,
            title=data.title,
        ),
    )


@ensure_execution.override(execution=MemberRemove, network="http")
async def remove_member_http(self: "OnebotProtocol", execution: MemberRemove) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.member)

    _check_execution(
        await self._http_post(
            "/set_group_kick",
            {
                "group_id": group_id,
                "user_id": user_id,
            },
        )
    )


@ensure_execution.override(execution=MemberRemove, network="ws")
async def remove_member_ws(self: "OnebotProtocol", execution: MemberRemove) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.member)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_kick",
            {
                "group_id": group_id,
                "user_id": user_id,
            },
        )
    )


@ensure_execution.override(execution=MemberMute, network="http")
async def mute_http(self: "OnebotProtocol", execution: MemberMute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._http_post(
            "/set_group_ban",
            {
                "group_id": group_id,
                "user_id": user_id,
                "duration": execution.duration,
            },
        )
    )


@ensure_execution.override(execution=MemberMute, network="ws")
async def mute_ws(self: "OnebotProtocol", execution: MemberMute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_ban",
            {
                "group_id": group_id,
                "user_id": user_id,
                "duration": execution.duration,
            },
        )
    )


@ensure_execution.override(execution=MemberUnmute, network="http")
async def unmute_http(self: "OnebotProtocol", execution: MemberUnmute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._http_post(
            "/set_group_ban",
            {
                "group_id": group_id,
                "user_id": user_id,
                "duration": 0,
            },
        )
    )


@ensure_execution.override(execution=MemberUnmute, network="ws")
async def unmute_ws(self: "OnebotProtocol", execution: MemberUnmute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_ban",
            {
                "group_id": group_id,
                "user_id": user_id,
                "duration": 0,
            },
        )
    )


@ensure_execution.override(execution=GroupMute, network="http")
async def group_mute_http(self: "OnebotProtocol", execution: GroupMute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._http_post(
            "/set_group_whole_ban",
            {
                "group_id": group_id,
                "enable": True,
            },
        )
    )


@ensure_execution.override(execution=GroupMute, network="ws")
async def group_mute_ws(self: "OnebotProtocol", execution: GroupUnmute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_whole_ban",
            {
                "group_id": group_id,
                "enable": True,
            },
        )
    )


@ensure_execution.override(execution=GroupUnmute, network="http")
async def group_unmute_http(self: "OnebotProtocol", execution: GroupUnmute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._http_post(
            "/set_group_whole_ban",
            {
                "group_id": group_id,
                "enable": False,
            },
        )
    )


@ensure_execution.override(execution=GroupUnmute, network="ws")
async def group_unmute_ws(self: "OnebotProtocol", execution: GroupUnmute) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_whole_ban",
            {
                "group_id": group_id,
                "enable": False,
            },
        )
    )


@ensure_execution.override(execution=MemberPromoteToAdministrator, network="http")
async def promote_to_admin_http(self, execution: MemberPromoteToAdministrator) -> None:
    _check_execution(
        await self._http_post(
            "/set_group_admin",
            {
                "group_id": isinstance(execution.group, Contactable)
                and execution.group.id
                or execution.group,
                "user_id": _extract_and_check_as_memberid(execution.target),
                "enable": True,
            },
        )
    )


@ensure_execution.override(execution=MemberPromoteToAdministrator, network="ws")
async def promote_to_admin_ws(self, execution: MemberPromoteToAdministrator) -> None:
    _check_execution(
        await self._ws_client_send_packet(
            "set_group_admin",
            {
                "group_id": execution.group,
                "user_id": _extract_and_check_as_memberid(execution.target),
                "enable": True,
            },
        )
    )


@ensure_execution.override(execution=MemberDemoteFromAdministrator, network="http")
async def demote_from_admin_http(self, execution: MemberDemoteFromAdministrator) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._http_post(
            "/set_group_admin",
            {
                "group_id": group_id,
                "user_id": user_id,
                "enable": False,
            },
        )
    )


@ensure_execution.override(execution=MemberDemoteFromAdministrator, network="ws")
async def demote_from_admin_ws(self, execution: MemberDemoteFromAdministrator) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_admin",
            {
                "group_id": group_id,
                "user_id": user_id,
                "enable": False,
            },
        )
    )


@ensure_execution.override(execution=MemberNicknameSet, network="http")
async def set_nickname_http(self: "OnebotProtocol", execution: MemberNicknameSet) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._http_post(
            "/set_group_card",
            {
                "group_id": group_id,
                "user_id": user_id,
                "card": execution.nickname,
            },
        )
    )


@ensure_execution.override(execution=MemberNicknameSet, network="ws")
async def set_nickname_ws(self: "OnebotProtocol", execution: MemberNicknameSet) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_card",
            {
                "group_id": group_id,
                "user_id": user_id,
                "card": execution.nickname,
            },
        )
    )


@ensure_execution.override(execution=MemberNicknameClear, network="http")
async def clear_nickname_http(self: "OnebotProtocol", execution: MemberNicknameClear) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._http_post(
            "/set_group_card",
            {
                "group_id": group_id,
                "user_id": user_id,
                "card": "",
            },
        )
    )


@ensure_execution.override(execution=MemberNicknameClear, network="ws")
async def clear_nickname_ws(self: "OnebotProtocol", execution: MemberNicknameClear) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    user_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_card",
            {
                "group_id": group_id,
                "user_id": user_id,
                "card": "",
            },
        )
    )


@ensure_execution.override(execution=GroupNameSet, network="http")
async def set_group_name_http(self: "OnebotProtocol", execution: GroupNameSet) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._http_post(
            "/set_group_name",
            {
                "group_id": group_id,
                "name": execution.name,
            },
        )
    )


@ensure_execution.override(execution=GroupNameSet, network="ws")
async def set_group_name_ws(self: "OnebotProtocol", execution: GroupNameSet) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_name",
            {
                "group_id": group_id,
                "name": execution.name,
            },
        )
    )


@ensure_execution.override(execution=GroupLeave, network="http")
async def leave_group_http(self: "OnebotProtocol", execution: GroupLeave) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._http_post(
            "/set_group_leave",
            {
                "group_id": group_id,
            },
        )
    )


@ensure_execution.override(execution=GroupLeave, network="ws")
async def leave_group_ws(self: "OnebotProtocol", execution: GroupLeave) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_leave",
            {
                "group_id": group_id,
            },
        )
    )


@ensure_execution.override(execution=MemberSpecialTitleSet, network="http")
async def set_special_title_http(self: "OnebotProtocol", execution: MemberSpecialTitleSet) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    target_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._http_post(
            "/set_group_special_title",
            {
                "group_id": group_id,
                "user_id": target_id,
                "title": execution.title,
            },
        )
    )


@ensure_execution.override(execution=MemberSpecialTitleSet, network="ws")
async def set_special_title_ws(self: "OnebotProtocol", execution: MemberSpecialTitleSet) -> None:
    group_id = _extract_and_check_as_groupid(execution.group)
    target_id = _extract_and_check_as_memberid(execution.target)

    _check_execution(
        await self._ws_client_send_packet(
            "set_group_special_title",
            {
                "group_id": group_id,
                "user_id": target_id,
                "title": execution.title,
            },
        )
    )


@ensure_execution.override(execution=MessageSend, network="http")
@ensure_execution.override(execution=MessageSend, network="ws")
async def send_message(self: "OnebotProtocol", execution: MessageSend) -> MessageId:
    using_method = None
    group_id = None
    friend_id = None

    target = ctx_target.get()

    if isinstance(target, str):
        raise ValueError("target as a target_id, must be a Contactable/ref")

    if isinstance(target, ref):
        if ref.type == "group":
            group_id = int(ref.id)
            using_method = "send_group_msg"
        elif ref.type == "friend":
            friend_id = int(ref.id)
            using_method = "send_private_msg"
        else:
            raise ValueError(f"unsupported ref type: {ref.type}")
    elif isinstance(target, Contactable):
        if isinstance(target.profile, MemberProfile):
            if target.profile.group is None:
                raise ValueError("target.profile.group is null")
            group_id = int(target.profile.group.id)
            using_method = "send_group_msg"
        elif isinstance(target.profile, FriendProfile):
            friend_id = int(target.id)
            using_method = "send_private_msg"
        else:
            raise ValueError(f"unsupported profile type: {type(target.profile)}")
    else:
        raise ValueError(f"unsupported target: {type(target)}")

    data = await self._onebot_send_packet(
        using_method,
        {
            **({"group_id": group_id} if group_id is not None else {}),
            **({"user_id": friend_id} if friend_id is not None else {}),
            "message": [
                *(await self.serialize_message(execution.message)),
                *(
                    [
                        {
                            "type": "reply",
                            "data": {
                                "id": execution.reply.id
                                if isinstance(execution.reply, MessageId)
                                else execution.reply
                            },
                        }
                    ]
                    if execution.reply
                    else []
                ),
            ],
        },
    )

    _check_execution(data)
    return MessageId(id=str(data["data"]["message_id"]))


@ensure_execution.override(execution=MessageRevoke, network="http")
async def revoke_message_http(self: "OnebotProtocol", execution: MessageRevoke) -> None:
    _check_execution(
        await self._http_post(
            "/delete_msg",
            {
                "message_id": int(
                    isinstance(execution.message_id, MessageId)
                    and execution.message_id.id
                    or execution.message_id
                ),
            },
        )
    )


@ensure_execution.override(execution=MessageRevoke, network="ws")
async def revoke_message_ws(self: "OnebotProtocol", execution: MessageRevoke) -> None:
    _check_execution(
        await self._ws_client_send_packet(
            "delete_msg",
            {
                "message_id": int(execution.message_id),
            },
        )
    )


@ensure_execution.override(execution=MessageFetch, network="http")
async def fetch_message_http(self: "OnebotProtocol", execution: MessageFetch) -> MessageFetchResult:
    return MessageFetchResult.parse_obj(
        _check_execution(
            await self._http_get(
                "/get_msg",
                {
                    "message_id": str(
                        execution.message_id.id
                        if isinstance(execution.message_id, MessageId)
                        else execution.message_id
                    )
                },
            )
        )
    )


@ensure_execution.override(execution=MessageFetch, network="ws")
async def fetch_message_ws(self: "OnebotProtocol", execution: MessageFetch) -> MessageFetchResult:
    return MessageFetchResult.parse_obj(
        _check_execution(
            await self._ws_client_send_packet(
                "get_msg",
                {
                    "message_id": str(
                        execution.message_id.id
                        if isinstance(execution.message_id, MessageId)
                        else execution.message_id
                    )
                },
            )
        )
    )


@ensure_execution.override(execution=MessageSendPrivate, network="http")
@ensure_execution.override(execution=MessageSendPrivate, network="ws")
async def send_private_message_ws(self, execution: MessageSendPrivate) -> MessageId:
    target = ctx_target.get()

    if isinstance(target, str):
        if target.isdigit():
            target_id = int(target)
        else:
            raise ValueError("target as a str, must be digit")

    else:
        if hasattr(target, "profile"):
            if isinstance(target, Contactable) and isinstance(target.profile, FriendProfile):
                target_id = int(target.id)
            else:
                raise ValueError("unsupported Contactable with profile, it need to be a `friend`.")
        else:
            raise ValueError("unsupported target without profile and not a str")

    data = await self._onebot_send_packet(
        "send_private_msg",
        {
            "user_id": target_id,
            "message": [
                *(await self.serialize_message(execution.message)),
                *(
                    [
                        {
                            "type": "reply",
                            "data": {
                                "id": execution.reply.id
                                if isinstance(execution.reply, MessageId)
                                else execution.reply
                            },
                        }
                    ]
                    if execution.reply
                    else []
                ),
            ],
        },
    )

    _check_execution(data)

    return MessageId(id=str(data["message_id"]))


@ensure_execution.override(execution=FetchAvatar, network="ws")
@ensure_execution.override(execution=FetchAvatar, network="http")
@ensure_execution.override(execution=FetchAvatar, network="http-service")
@ensure_execution.override(execution=FetchAvatar, network="ws-service")
async def fetch_avatar_http(self: "OnebotProtocol", execution: FetchAvatar) -> AvatarResource:
    if isinstance(execution.target, Contactable):
        if isinstance(execution.target.profile, GroupProfile):
            http_client = self.avilla.network_interface.get_by_class(AbstractHttpClient)
            return AvatarResource(
                RawProvider(
                    (
                        await http_client.get(
                            URL(
                                f"https://p.qlogo.cn/gh/{execution.target.id}/{execution.target.id}/0"
                            )
                        )
                    ).transform()
                )
            )
        elif isinstance(execution.target.profile, (FriendProfile, StrangerProfile)):
            http_client = self.avilla.network_interface.get_by_class(AbstractHttpClient)
            return AvatarResource(
                RawProvider(
                    (
                        await http_client.get(
                            URL(f"https://q1.qlogo.cn/g?b=qq&nk={execution.target.id}&s=640")
                        )
                    ).transform()
                )
            )
        else:
            raise ValueError("unsupported profile")
    elif isinstance(execution.target, ref):
        if execution.target.type == "group":
            http_client = self.avilla.network_interface.get_by_class(AbstractHttpClient)
            return AvatarResource(
                RawProvider(
                    (
                        await http_client.get(
                            URL(
                                f"https://p.qlogo.cn/gh/{execution.target.id}/{execution.target.id}/0"
                            )
                        )
                    ).transform()
                )
            )
        elif execution.target.type in {"friend", "stranger"}:
            http_client = self.avilla.network_interface.get_by_class(AbstractHttpClient)
            return AvatarResource(
                RawProvider(
                    (
                        await http_client.get(
                            URL(f"https://q1.qlogo.cn/g?b=qq&nk={execution.target.id}&s=640")
                        )
                    ).transform()
                )
            )
        else:
            raise ValueError("unsupported ref type")
    else:
        raise ValueError("unsupported target")
